#include "fileio.h"

bool write_EOF(FILE *f) {
 return true; }
bool read_EOF(FILE *f) {
 return fgetc(f) == EOF; }

bool write_byte(uint8_t byte, FILE *f) {
 return fputc(byte, f) != EOF; }
bool read_byte(uint8_t *byte, FILE *f) {
 int c = fgetc(f);
 if (c != EOF) *byte = c;
 return c != EOF; }
bool check_byte(const char *msg, uint8_t byte, FILE *f) {
 uint8_t b = 0;
 if (!read_byte(&b, f) || b != byte)
  return ERROR("Check for '%s' failed, expected %u, got %u."
              , msg, byte, b), false;
 return true; }

bool write_bool(bool boolean, FILE *f) {
 return write_byte((uint8_t) boolean, f); }
bool read_bool(bool *boolean, FILE *f) {
 uint8_t byte;
 if (!read_byte(&byte, f) || !(byte == 0 || byte == 1)) return false;
 return *boolean = (bool) byte, true; }

bool write_str(char *str, FILE *f) {
 for (; *str; str++)
  if (!write_byte((uint8_t) *str, f)) return false;
 return true; }
bool check_str(const char *msg, char *str, FILE *f) {
 for (; *str; str++)
  if (!check_byte("check_*str", (uint8_t) *str, f))
   return ERROR("Check for '%s' failed.", msg), false;
 return true; }

bool read_nstr(size_t n, char **str, FILE *f) {
 *str = malloc((n+1) * sizeof **str);
 if (!*str) return MALLOC_ERROR(), false;
 for (size_t j = 0; j < n; j++)
  if (!read_byte((uint8_t *) &(*str)[j], f)) return false;
 (*str)[n] = '\0';
 return true; }

#define WRITE_TYPE(T, t, f)  \
 for (int j = 0; j < sizeof (T); j++) { \
  if (!write_byte((uint8_t) (t & 0xFF), f)) return false; \
  t >>= 8; } \
 return true;
#define READ_TYPE(T, pt, f)  \
 *pt = 0; \
 uint8_t b; \
 for (int j = 0; j < sizeof (T); j++) { \
  if (!read_byte(&b, f)) return false; \
  *pt |= (T) b << (8*j); } \
 return true;

bool write_size_t(size_t size, FILE *f) {
 WRITE_TYPE(size_t, size, f) }
bool read_size_t(size_t *size, FILE *f) {
 READ_TYPE(size_t, size, f); }

bool write_uint64_t(uint64_t n, FILE *f) {
 WRITE_TYPE(uint64_t, n, f) }
bool read_uint64_t(uint64_t *n, FILE *f) {
 READ_TYPE(uint64_t, n, f) }

bool write_xit_t(xit_t x, FILE *f) {
 WRITE_TYPE(xit_t, x, f) }
bool read_xit_t(xit_t *x, FILE *f) {
 READ_TYPE(xit_t, x, f) }

#undef WRITE_TYPE
#undef READ_TYPE

#define WRITE_TYPES(write, size, ts, f) \
 for (size_t j = 0; j < size; j++) \
  if (!write(ts[j], f)) return false; \
 return true;
#define READ_TYPES(read, size, pts, f) \
 *pts = malloc(size * sizeof **pts); \
 if (!*pts) return MALLOC_ERROR(), false; \
 for (size_t j = 0; j < size; j++) \
  if (!read(&(*pts)[j], f)) return false; \
 return true;

#define READ_TYPES_OVERALLOCATING(read, maxsize, size, pts, f) \
 *pts = calloc(maxsize, sizeof **pts); \
 if (!*pts) return MALLOC_ERROR(), false; \
 for (size_t j = 0; j < size; j++) \
  if (!read(&(*pts)[j], f)) return false; \
 return true;

bool write_uint64_ts(size_t size, uint64_t *ns, FILE *f) {
 WRITE_TYPES(write_uint64_t, size, ns, f) }
bool read_uint64_ts(size_t size, uint64_t **ns, FILE *f) {
 READ_TYPES(read_uint64_t, size, ns, f) }

bool write_xit_ts(size_t size, xit_t *xits, FILE *f) {
 WRITE_TYPES(write_xit_t, size, xits, f) }
bool read_xit_ts(size_t size, xit_t **xits, FILE *f) {
 READ_TYPES(read_xit_t, size, xits, f) }
bool read_xit_ts_overallocating(
 size_t maxsize, size_t size, xit_t **xits, FILE *f
) {
 READ_TYPES_OVERALLOCATING(read_xit_t, maxsize, size, xits, f) }

bool write_natural_t(natural_t *nat, FILE *f) {
 return write_xit_t(nat->radix, f)
     && write_xit_t(nat->base, f)
     && write_size_t(nat->n_xits_max, f)
     && write_size_t(nat->n_xits, f)
     && write_xit_ts(nat->n_xits, nat->xits, f)
     && write_bool(nat->overflowed, f)
     && write_bool(nat->sequence_member, f)
     && write_uint64_t(nat->time, f); }
bool read_natural_t(natural_t *nat, FILE *f) {
 return read_xit_t(&nat->radix, f)
     && read_xit_t(&nat->base, f)
     && read_size_t(&nat->n_xits_max, f)
     && read_size_t(&nat->n_xits, f)
     && read_xit_ts_overallocating(nat->n_xits_max, nat->n_xits, &nat->xits, f)
     && read_bool(&nat->overflowed, f)
     && read_bool(&nat->sequence_member, f)
     && read_uint64_t(&nat->time, f); }

bool write_natural_ts(size_t size, natural_t *nats, FILE *f) {
 WRITE_TYPES(write_natural_t, size, &nats, f) }
bool read_natural_ts(size_t size, natural_t **nats, FILE *f) {
 READ_TYPES(read_natural_t, size, nats, f) }

#undef WRITE_TYPES
#undef READ_TYPES
#undef READ_TYPES_OVERALLOCATING

bool write_computation(char *filename, computation_t *computation) {
 FILE *f = fopen(filename, "wb");
 if (!f) return FILE_WRITE_ERROR(filename), false;

 size_t comment_size = str_len(computation->comment);
 return write_str(MAGIC_NUMBER, f)
     && write_byte((uint8_t) sizeof (size_t), f)
     && write_byte((uint8_t) sizeof (xit_t), f)
     && write_byte((uint8_t) sizeof (xit_carry_t), f)
     && write_byte((uint8_t) sizeof (rit_count_t), f)
     && write_size_t(comment_size, f)
     && write_str(computation->comment, f)
     && write_size_t(computation->n_terms, f)
     && write_uint64_ts(computation->n_terms, computation->ns, f)
     && write_natural_ts(computation->n_terms, computation->two_pow_ns, f)
     && write_EOF(f)
     && (fclose(f), true); }

bool read_computation(char *filename, computation_t *computation) {
 FILE *f = fopen(filename, "rb");
 if (!f) return FILE_READ_ERROR(filename), false;

 size_t comment_size;
 return check_str("magic number", MAGIC_NUMBER, f)
     && check_byte("sizeof (size_t)", (uint8_t) sizeof (size_t), f)
     && check_byte("sizeof (xit_t)", (uint8_t) sizeof (xit_t), f)
     && check_byte("sizeof (xit_carry_t)", (uint8_t) sizeof (xit_carry_t), f)
     && check_byte("sizeof (rit_count_t)", (uint8_t) sizeof (rit_count_t), f)
     && read_size_t(&comment_size, f)
     && read_nstr(comment_size, &computation->comment, f)
     && read_size_t(&computation->n_terms, f)
     && read_uint64_ts(computation->n_terms, &computation->ns, f)
     && read_natural_ts(computation->n_terms, &computation->two_pow_ns, f)
     && read_EOF(f)
     && (fclose(f), true); }


