#include "pargs.h"

void print_help() {
 #define P printf
 P("Extending A056154 -- Jonathan Frech, September and October of 2019\n\n");
 P("Usage: a056154 [mode] [options] [filename]\n\n");
 P("Modes and options:\n");
 P(" -l, --log.................: Print log messages.\n\n");
 P(" -h, --help................: Help mode. Print this message and exit.\n\n");
 P(" -n, --new.................: New mode. Create a new data file.\n");
 P(" --3^{pow}.................: Set base to `3^pow`, where `pow > 0`.\n");
 P(" --comment={comment}.......: Set file comment to `comment`.\n\n");
 P(" -i, --info................: Information mode. Print information about\n"
   "                             given file. Set by default.\n");
 P("  --as_ternary={n}.........: Print `2^n`'s ternary representation.\n");
 P("  --as_ternary=1+{n}.......: Print `2^(1+n)`'s ternary representation.\n");
 P(" -c, --compute.............: Computation mode. Compute on given file.\n");
 P("  --target={target}........: Search n up to but not including `target`.\n");
 P("  --target=+{+target}......: Search for `+target` further `n`.\n");
 P("  --target=max.............: Set search for n up to but not including\n"
   "                             `max_target=%" PRIu64 "`.\n", max_target);
 P("  --one-pass...............: Compute using only one pass. May result in\n"
   "                             performance benefits.\n");
 P("  --nosave.................: Do not save computational progress.\n");
 P("  --print_time.............: Print time used to compute (in seconds).\n");
 P("  --only_verify_rit_count_t: Only verify that `sizeof (rit_count_t)`\n"
   "                             is compatible with the given target.\n"
   "                             No computation is performed.\n\n");
 P(" -e, --edit................: Edit mode. Alter the given file.\n");
 P("  --remove_non_members.....: Remove any saved representations of `2^n`\n"
   "                             (except the currently active) which are not\n"
   "                             of sequence members `n`.\n");
 P("  --new_comment={comment}..: Update the file's comment.\n");
 #undef P
}

bool parse_args(int argc, char **argv, pargs_t *pargs) {
 pargs->filename = NULL;
 pargs->log = false;

 pargs->mode_help = false;

 pargs->mode_new = false;
 pargs->three_pow = 0;
 pargs->comment = NULL;

 pargs->mode_info = false;
 pargs->as_ternary = 0;
 pargs->as_ternary_plus_one = false;

 pargs->mode_compute = false;
 pargs->target = 0;
 pargs->target_plus = false;
 pargs->nosave = false;
 pargs->print_time = false;
 pargs->only_verify_rit_count_t = false;
 pargs->one_pass = false;

 pargs->mode_edit = false;
 pargs->remove_non_members = false;
 pargs->new_comment = NULL;

 if (argc <= 1)
  return ERROR("No arguments given. Use '--help' for help."), false;

 #define NO_MODE_SET \
  (!pargs->mode_help && !pargs->mode_new && !pargs->mode_info \
  && !pargs->mode_compute && !pargs->mode_edit)

 char *argument_inducing_mode = NULL;
 char *arg = NULL;

 #define SET_MODE(MODE) \
   if (NO_MODE_SET) pargs->MODE = true, argument_inducing_mode = arg; \
   if (!pargs->MODE) \
    return ERROR("Mode (set by %s) clash, %s.", argument_inducing_mode, arg) \
         , false;

 bool interpret_args = true;
 for (int a = 1; a < argc; a++) {
  arg = argv[a];
  bool is_filename = false;
  if (!interpret_args) is_filename = true;
  else {
   if (str_equal(arg, "--")) interpret_args = false;

   else if (str_prefix(arg, "--")) {
    /**/ if (str_equal(arg, "--log")) pargs->log = true;

    else if (str_equal(arg, "--help")) { SET_MODE(mode_help); }

    else if (str_equal(arg, "--new")) { SET_MODE(mode_new); }
    else if (str_prefix(arg, "--3^")) {
     SET_MODE(mode_new);
     if (pargs->three_pow)
      return ERROR("3^-base already specified as '%" PRIu64 "', cannot be "
                   "specified as '%s'.", pargs->three_pow, arg), false;
     uint64_t three_pow = (uint64_t) atoll(str_len("--3^") + arg);
     if (three_pow <= 0)
      return ERROR("Invalid 3^-base '%s'.", arg), false;
     pargs->three_pow = three_pow; }
    else if (str_prefix(arg, "--comment=")) {
     SET_MODE(mode_new); pargs->comment = str_len("--comment=") + arg; }

    else if (str_equal(arg, "--info")) { SET_MODE(mode_info); }
    else if (str_prefix(arg, "--as_ternary=")) {
     SET_MODE(mode_info);
     if (pargs->as_ternary)
      return ERROR("Already showing %" PRIu64 " as ternary, cannot also "
                   "specify '%s'.", pargs->as_ternary, arg), false;
     uint64_t as_ternary_value = 0;
     if (str_prefix(arg, "--as_ternary=1+")) {
      as_ternary_value = (uint64_t) atoll(str_len("--as_ternary=1+") + arg);
      pargs->as_ternary_plus_one = true; }
     else {
      as_ternary_value = (uint64_t) atoll(str_len("--as_ternary=") + arg);
      pargs->as_ternary_plus_one = false; }
     if (!as_ternary_value) return ERROR("Invalid argument '%s'.", arg), false;
     pargs->as_ternary = as_ternary_value; }

    else if (str_equal(arg, "--compute")) { SET_MODE(mode_compute); }
    else if (str_prefix(arg, "--target=")) {
     SET_MODE(mode_compute);
     if (pargs->target)
      return ERROR("Target already specified as '%" PRIu64 "', cannot be "
                   "specified as '%s'.", pargs->target, arg), false;
     uint64_t target_value;
     if (str_equal(arg, "--target=max")) target_value = max_target;
     else if (str_prefix(arg, "--target=+")) {
      pargs->target_plus = true;
      target_value = (uint64_t) atoll(str_len("--target=+") + arg); }
     else target_value = (uint64_t) atoll(str_len("--target=") + arg);
     if (target_value <= 0)
      return ERROR("Invalid target '%s'.", arg), false;
     pargs->target = target_value; }
    else if (str_equal(arg, "--nosave")) {
     SET_MODE(mode_compute); pargs->nosave = true; }
    else if (str_equal(arg, "--print_time")) {
     SET_MODE(mode_compute); pargs->print_time = true; }
    else if (str_equal(arg, "--only_verify_rit_count_t")) {
     SET_MODE(mode_compute); pargs->only_verify_rit_count_t = true; }
    else if (str_equal(arg, "--one-pass")) {
     SET_MODE(mode_compute); pargs->one_pass = true; }

    else if (str_equal(arg, "--edit")) { SET_MODE(mode_edit); }
    else if (str_equal(arg, "--remove_non_members")) {
     SET_MODE(mode_edit); pargs->remove_non_members = true; }
    else if (str_prefix(arg, "--new_comment=")) {
     SET_MODE(mode_edit);
     pargs->new_comment = str_len("--new_comment=") + arg; }

    else return ERROR("Unknown argument '%s'.", arg), false; }

   else if (str_prefix(arg, "-")) {
    for (char c; (c = *++arg); )
     /**/ if (c == 'l') pargs->log = true;
     else if (c == 'h') { SET_MODE(mode_help); }
     else if (c == 'n') { SET_MODE(mode_new); }
     else if (c == 'i') { SET_MODE(mode_info); }
     else if (c == 'c') { SET_MODE(mode_compute); }
     else if (c == 'e') { SET_MODE(mode_edit); }
     else return ERROR("Unknown single-character argument '-%c' in argument "
                       "'%s'.", c, arg), false; }

   else is_filename = true;
  }
  if (is_filename) {
   if (pargs->filename)
    return ERROR("Filename already specified as '%s', cannot be specified as "
                 "'%s'.", pargs->filename, arg), false;
   else pargs->filename = arg; }
 }

 arg = NULL;
 if (NO_MODE_SET) { SET_MODE(mode_info); }
 if (!pargs->mode_help && !pargs->filename)
  return ERROR("No filename specified."), false;
 if (!pargs->target) pargs->target = max_target;
 if (!pargs->three_pow) pargs->three_pow = default_three_pow;

 if (pargs->mode_edit && !pargs->remove_non_members && !pargs->new_comment)
  return ERROR("Nothing to edit."), false;

 #undef NO_MODE_SET
 #undef SET_MODE

 return true; }
